#!/usr/bin/env python3
"""
An Interesting Consequence of Mathematics
by Javantea
April 7, 2015

What RC4 key produces the same first four bytes of keystream as another key?
It should only take testing 2-4 billion keys to find a single instance of this, 
so let's do it.

Umm it took seconds to find a thousand collisions.

"""
import Crypto.Cipher.ARC4
import sys
import struct
import Crypto.Util.number
import random
import Crypto.Util.randpool

def collideKeyStream(keystreams, keys=0x100000, start = 0, length = 64):
	if length != 64 and length != 128:
		raise ValueError("Length must be 64 or 128.")
	#end if
	output = []
	i = start
	while len(output) < keys:
		if length == 64:
			# 64-bit key.
			key = struct.pack('>Q', i)
		elif length == 128:
			# TODO: 128-bit key.
			key = struct.pack('>Q', i >> 64) + struct.pack('>Q', i & 0xffffffffffffffff)
		#end if
		keystream = Crypto.Cipher.ARC4.new(key).encrypt(b'\x00\x00\x00\x00')
		if keystream in keystreams:
			output.append(key)
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output, i
#end def collideKeyStream(keystreams, keys=0x100000, start = 0, length = 64)

def collideRandomKeyStream(keystreams, keys=0x100000, length = 64):
	if length != 64 and length != 128:
		raise ValueError("Length must be 64 or 128.")
	#end if
	rand = Crypto.Util.randpool.RandomPool()
	output = []
	i = 0
	while len(output) < keys:
		key = rand.get_bytes(length>>3)
		if len(key) != (length >> 3):
			# This will happen 1 in ... times.
			#if (length >> 3) - len(key) > 1:
			#	raise Exception("Serious lack of randomness crash %i" % (len(key)))
			#end if
			# This will happen 1 in 256 times.
			key = (b'\x00' * ((length >> 3) - len(key))) + key
		#end if
		keystream = Crypto.Cipher.ARC4.new(key).encrypt(b'\x00\x00\x00\x00')
		if keystream in keystreams:
			output.append(key)
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output, i
#end def collideRandomKeyStream(keystreams, keys=0x100000, length = 64)

def getKeyStreams(start = 0, keystreams=0x100000, length=64):
	if length != 64 and length != 128:
		raise ValueError("Length must be 64 or 128.")
	#end if
	output = []
	i = start
	for v in range(keystreams):
		if length == 64:
			# 64-bit key.
			key = struct.pack('>Q', i)
		elif length == 128:
			# TODO: 128-bit key.
			key = struct.pack('>Q', i >> 64) + struct.pack('>Q', i & 0xffffffffffffffff)
		#end if
		keystream = Crypto.Cipher.ARC4.new(key).encrypt(b'\x00\x00\x00\x00')
		output.append(keystream)
		i += 1
		if i & 0xffffff == 0: print(i)
	#loop
	return output
#end def getKeyStreams(start = 0, keystreams=0x100000, length=64)

def main():
	start = 0x100000
	randomize = False
	keys_we_want = 1024
	if '-r' in sys.argv:
		randomize = True
		sys.argv.pop(sys.argv.index('-r'))
	#end if
	if len(sys.argv) > 1:
		start = int(sys.argv[1])
		if len(sys.argv) > 2:
			keys_we_want = int(sys.argv[2])
		#end if
	#end if
	print("Starting at:", start)
	if start < 0x100000:
		print("This isn't going to work. start should be greater than", 0x100000)
	#end if
	
	tons_of_keystreams = getKeyStreams(0, 0x100000)
	tons_of_keystreams_set = set(tons_of_keystreams)
	print('len(tons_of_keystreams_set):', len(tons_of_keystreams_set), '%3.3f' % (len(tons_of_keystreams_set) / 0x100000))
	
	if randomize:
		start = 0
		keys_out, end = collideRandomKeyStream(tons_of_keystreams_set, keys_we_want)
	else:
		keys_out, end = collideKeyStream(tons_of_keystreams_set, keys_we_want, start)
	#end if
	for key in keys_out:
		keystream = Crypto.Cipher.ARC4.new(key).encrypt(b'\x00\x00\x00\x00')
		print(key, keystream, tons_of_keystreams.index(keystream))
	#next key
	print(len(keys_out), "keys found in", end-start, "attempts")
#end def main()

if __name__ == '__main__':
	main()
#end if

